"""
This program was written jointly by Vasily Dolgushev and Geoffrey Schneider

Tools for working with Gerstenhaber words are given here. 


"""



# coding: utf-8
import itertools
import time
import math

from copy import deepcopy as deepcopy

from math import factorial as fact # this is for testing

from itertools import combinations as comb

# for example comb(tup, 2) returns the generator "tup choose 2"

from LinCombGraphs import Simplify as Simplify
from LinCombGraphs import binomC as binomC 
# binomC is for testing
from LinCombGraphs import mult as mult
from LinCombGraphs import Part as Part
from LinCombGraphs import Perm as Perm
from LinCombGraphs import linExt as linExt


"""
NoRpn(L) returns True is L does not have repetitions
Otherwise it returns False.
Note that L can be a list or a tuple.
We need the function NoRpn for testing.
"""

#Tested
def NoRpn(L):
    for el in L:
        if L.count(el)>1:
            return False
    return True


#
# FUNCTIONS FOR LINEAR COMBINATIONS.
# OUR LINEAR COMBINATIONS ARE COMBINATIONS OF TUPLES OR LISTS
#



# Here are examples of a Ger word:


aa =[(1,),(2,)] # this is a_1 a_2

brack = [(1,2)] # this is \{a_1, a_2\} 

W = [(1,),(2,3)]

Sl = [(2,4),(1,3)]

WW =  [(1,4), (3,2,5)]
    
G = [(1,2),(3,4),(5,)]
GG = [(5,),(1,2),(3,4)]
GGG = [(3,4),(5,), (1,2)]


"""
linear combinations of Ger words are
represnted as nested lists [[coeff1, W1], [coeff2, W2]]
"""

# Examples of LaLie words:
L2  = [[1,(1,2)]]
L3  = [[1,(1,2,3)]], 
LL3  = [[1,(2,1,3)]], 
L4  = [[1,(1,2,3,4)]], 
LL4  = [[1,(2,1,3,4)]]


"""
degG computes the degree of the Ger word W
Note that we need degree mod 2
"""

#Tested
def degG(W):
    tup = tuple(len(el)-1 for el in W)
    return sum(tup)

    

"""
lenG computes the length of the Ger word W:
"""
#Tested
def lenG(W):
    tup = tuple(len(el) for el in W)
    return sum(tup)

    

"""
multGer multiplies two linear combinations of Ger words.
Note that we do not sort and we do not simplify.
"""

#nothing to test
def multGer(W1,W2):
    Out = []
    for el in W1:
        for ent in W2:
            Out.append([el[0]*ent[0], el[1] + ent[1]])
    
    return Out

"""
SignG computes the sign the of the permutation of Ger word W into 
a Ger word WW
W and WW must have the same length and the SAME Lie words
In other words, WW = Sign (W,WW) W.
"""

# Tested
def SignG(W,WW):
    s = 1
    for t in comb(range(0,len(W)),2):
        if WW.index(W[t[0]]) > WW.index(W[t[1]]):
            fac = (-1)**((len(W[t[0]])-1)*(len(W[t[1]])-1))
            s=s*fac
                
    return s


"""
ActG(sig,W) returns the left action of the permutation sig on a Ger word W
sig must be an element of S_{len(W)}
NOTE that the output of Act is usually a non-standard Ger word.
See the function toStanGer below
"""

#Tested
def ActG(sig,W):
    out = []
    for el in W:
        out.append(tuple(sig[i-1] for i in el))
        
    return out

"""
ActL(sig,L) returns the result of the action of a permutatio sig on
a LaLie word L (viewed as a tuple)
"""
#Tested
def ActL(sig,L):
    return tuple(sig[i-1] for i in L)

"""
lcActL is the linear extension of ActL.
I.e. lcActL can be applied to linear combinations of 
LaLie words. 
"""

#Tested
def lcActL(sig, lc):
    def fun(t):
        return [[1, ActL(sig,t)]]
    return linExt(fun,lc)


    
"""
shiftL produces a new tuple from a numerical tuple tup by shifting all
entries j > i up by q
"""

#Test
def shiftL(tup,i,q):
    out = ()
    for j in tup:
        if j <= i:
            out = out + (j,)
        else:
            out = out + (j+q,)
            
    return out


"""
Given a tuple (a_1,a_2, ..., a_k) with k > 1 "split" returns the
linear combination - (a_1,(a_2,...a_k)) - (-1)^k ((a_2,...a_k), a_1)
if k > 2.
If k = 2, we get -(a_1, a_2) - (a_2, a_1).


Note the the degree of a tuple tup (viewes as the LaLie word)
has degree len(tup)-1 (mod 2)
"""
#Tested
def split(tup):
    k = len(tup)
    if k > 2:
        smtup = tup[1:k+1]
    else:
        smtup = tup[1]

    Out =[]
    Out.append( [-1, ( tup[0], smtup )] )
    Out.append( [(-1)**(k-1), (smtup, tup[0])] )
    return Out

"""
exactly 1 element of tup is a tuple. All other entries are integers.
ind_tup determines the index of this poor element
"""

# nothing to test
def ind_tup(tup):
    for el in tup:
        if type(el) == tuple:
            return tup.index(el)

"""
L is a linear combination of "associative words". 
Each associative word is a tuple of len > = 1 in which one element is a tuple 
and all other elements (if any) are integers.
bSlplit applied split to this sub-tuple and then concatenates.  
"""
    
#Tested
def bSplit(L):
    Out = []
    for el in L:
        ind = ind_tup(el[1])
        beg = el[1][0:ind] # the stuff preceding el[1][ind]
        end = el[1][ind+1:len(el[1])] # the stuff after el[1][ind]
        aux = split(el[1][ind]) # we splitted the subtuple
        for ent in aux:
            term = [el[0]*ent[0], beg + ent[1] + end] 
            Out.append(term[:])

    return Out


"""
fullSplit takes a tuple L (viewed as the LaLie word)
and produces the linear combination of the corresponding "associative words".
The output of fullSplit is a nested list [[coeff, stuff], ...]
NOTE: fullSplit CAN be applied to a tuple of length 1.
"""

#Tested
def fullSplit(L):
    if len(L) == 1:
        return [[1,L]]
    else:
        Aux = [[1,(L,)]]
        for i in range(len(L)-1):
            Out = bSplit(Aux)
            Aux = Out[:]
    
    return Out

"""
InsLLie computes the insertions of the LaLie word L into
the i-th slot of the LaLie word W. Both L and W are tuples.
The output is a linear combination. So it is a nested list.
"""
#Tested
def InsLLie(W,i,L):
    if len(L) == 1:
        return [ [1,W] ]
    if i != W[len(W)-1]:
        shW = shiftL(W,i, len(L)-1) # this is the case when i is not in the right most slot
        beg = shW[0:shW.index(i)]
        end = shW[shW.index(i)+1:len(shW)]
        shL = shiftL(L,0,i-1)
        elemA = fullSplit(shL)
        return [ [(-1)**((len(end)-1)*(len(L)-1)) * el[0], beg + el[1] + end ] for el in elemA]
    else:
        shL = shiftL(L,0,i-1)
        shW = shiftL(W,i, len(L)-1)
        return [ [1, shW[0:len(W)-1] + shL] ] # taking care of the insertion in the right most slot.


"""
toStandard takes a tuple (viewed) as the LaLie word and turns it
into the linear combination of standard Ger word of the form [( )]
toStandard CAN be applied to a standard LaLie word as well.
"""

#Tested
def toStandard(L):
    ind = L.index(max(L))
    if ind == len(L)-1:
        return [ [1, [L]] ]
    else:
        right = L[ind+1:len(L)] # after max(L)
        left = L[0:ind] # excluding the maximal element 
        return [ [el[0], [left + el[1] + (L[ind],)] ] for el in fullSplit(right)]


"""
toStanGer takes a Ger word W and returns the corresponding linear
combination of standard Ger words.

In the first part, we produce the linear combination of
products of standard LaLie words. In the second part we sort
the LaLie words with appropriate sign factors.
"""

#Tested
def toStanGer(W):
    Aux = [toStandard(tup) for tup in W]
    Temp = Aux[0]
    for i in range(1, len(W)):
        Temp = multGer(Temp,Aux[i])

    Out = []
    for el in Temp:
        G = el[1][:]
        G.sort()
        Out.append([el[0]*SignG(el[1],G), G[:]])

    return Out

"""
toStanlc takes a linear combination of Ger words and produces
the corresponding linear combination of standard Ger words. We do not apply Simplify here!
"""
#Tested
def toStanlc(G):
    Out =[]
    for El in G:
        Out = Out + mult(El[0], toStanGer(El[1]))
    
    return Out



"""
degBefore(W,i) is the sum of degrees of all LaLie words in a Ger word W from the first one till the (i-1)-th.
len(W) must be > i-1
"""

#Tested
def degBefore(W,i):
    tup = tuple(len(W[j])-1 for j in range(0,i-1))
    return sum(tup)

    """
Similarly degAfter
"""
#Tested
def degAfter(W,i):
    tup = tuple(len(W[j])-1 for j in range(i,len(W)))
    return sum(tup)


"""
Der prepares a linear combination of Ger words W for a derivation.
The output is a nested list of lists [coeff, X_i, [stuff without X_i] ]
The sign factor comes from the usual jumping.
Der can only be applied to a lin comb of Ger words with at least 1 multiplication!!!
"""
#Tested
def Der(W):
    Out = []
    for El in W:
        Arg =deepcopy(El[1])
        for i in range(1, len(Arg)+1):
            s = El[0]*(-1)**((len(Arg[i-1])-1)*degBefore(Arg,i))
            Out.append([s, Arg[i-1], Arg[0:i-1]+Arg[i:len(Arg)]]) 
    
    return Out


"""
Ins1 substitutes a Ger word W into the first slot of the LaLie word L.
The sing factor is the result of el[1] jumping over len(L)-2 brackets
NO shifts in this function!!!
"""

#Tested
def Ins1(L,W):
    Out = []
    for el in Der([[1,W]]):
        aux = fullSplit(el[1])
        for ent in aux:
            term = [el[0]*ent[0]*((-1)**( (len(el[1])-1)*len(L))), [ ent[1]+L[1:len(L)] ]+ el[2]]
            Out.append(deepcopy(term))

    return Out



"""
Adj applies {a_i,~} to a linear combination of Ger words W.
Just as for Ins1, no shifts in this function.
"""

#Tested
def Adj(i,W):
    return [ [el[0], [ (i,) + el[1] ] + el[2]] for el in Der(W)]



"""
InsLG substitutes a Ger word W into the i-th slot of the LaLie word L
No shifts here!
"""
#Tested
def InsLG(L,i,W):
    if len(L) == 1:
        return [[1,W[:]]]
    ind = L.index(i)
    if ind == len(L)-1:
        # this is the case when a_i is the right most elem of L
        Out = [[1,W[:]]]
        for j in range(len(L)-2,-1,-1):
            Out = Adj(L[j],Out)
        return Out
    if ind == 0:
        # this is the case when a_i is the left most elem of L
        return Ins1(L,W)
    # otherwise:
    Out = Ins1(L[ind:len(L)], W)
    for j in range(ind-1,-1,-1):
        Out = Adj(L[j],Out)
    return Out


"""
LL(W,i) returns the LaLie words with a_i
"""
#Tested
def LL(W,i):
    for el in W:
        if i in el:
            return el

"""
InsGG computes the insertion of the Ger word WW into the i-th slot of the Ger word W
Now we have the necessary shifts
The output is a linear combination of standard Ger words
NOTE: When we preform insertions, of standard Ger words, LaLie parts are
all standard. 
"""

#Tested
def InsGG(W,i,WW):
    G = [shiftL(el,i, lenG(WW)-1) for el in W] # G is obtained from W by shifting all j >i up by lenG(WW)-1
    GG = [shiftL(el,0, i-1) for el in WW] # shifted all j in WW up by i-1
    L = LL(G,i)
    ind = G.index(L) + 1 # this is the adjusted index of L in G
    LG = mult((-1)**(degG(GG)*degAfter(G,ind)) , InsLG(L,i,GG))
    right = multGer(LG,[[1,G[ind:len(G)]]]) # multiply LG with what stands to the right of L
    return toStanlc( multGer([[1,G[0:ind-1]]], right) )














